# Copyright (C) 2024 Peter Ropač
# This file is part of SPOMSO.
# SPOMSO is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
# SPOMSO is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.
# You should have received a copy of the GNU Lesser General Public License along with SPOMSO. If not, see <https://www.gnu.org/licenses/>.

import numpy as np
from typing import Callable
from spomso.cores.geom import GenericGeometry
from spomso.cores.transformations import EuclideanTransform
from spomso.cores.modifications import ModifyObject
from spomso.cores.sdf_2D import sdf_circle, sdf_box_2d, sdf_segment_2d, sdf_triangle_2d, sdf_rounded_box_2d
from spomso.cores.sdf_2D import sdf_sector, sdf_inf_sector, sdf_sector_old, sdf_ngon, sdf_arc
from spomso.cores.sdf_2D import sdf_parametric_curve_2d, sdf_segmented_curve_2d, sdf_segmented_line_2d
from spomso.cores.sdf_2D import sdf_point_cloud_2d


class GenericGeometry2D(EuclideanTransform, ModifyObject):

    def __init__(self, geo_sdf, *geo_parameters):
        EuclideanTransform.__init__(self)
        ModifyObject.__init__(self, geo_sdf)
        self.geo_parameters = geo_parameters
        self._sdf = geo_sdf

    def create(self, co):
        self._sdf = self.modified_object
        return self.apply(self._sdf, co, self.geo_parameters)

    def propagate(self, co, *parameters_):
        self._sdf = self.modified_object
        return self.apply(self._sdf, co, self.geo_parameters)


class Circle(GenericGeometry):
    """Circle defined by its radius.

    Args:
        radius: Radius of the circle.
    """

    def __init__(self, radius: float | int):
        GenericGeometry.__init__(self, sdf_circle, radius)
        self._radius = radius

    @property
    def radius(self) -> float | int:
        """Radius of the circle."""
        return self._radius


class NGon(GenericGeometry):
    """
    N-sided regular polygon, defined by the outer radius and the number of sides.

    Args:
        radius: Outer radius of the regular polygon.
        n_sides: Number of sides of the regular polygon.
    """

    def __init__(self, radius: float | int, n_sides: int):
        GenericGeometry.__init__(self, sdf_ngon, radius, n_sides)
        self._radius = radius
        self._n_sides = n_sides

    @property
    def radius(self) -> float | int:
        """Radius of the n-gon"""
        return self._radius

    @property
    def n_sides(self) -> int:
        """Number of sides in the n-gon"""
        return self._n_sides


class Rectangle(GenericGeometry):
    """
    Rectangle defined by its side lengths.

    Args:
        a: Side length along the x-axis.
        b: Side length along the y-axis.
    """

    def __init__(self, a: float | int, b: float | int):
        GenericGeometry.__init__(self, sdf_box_2d, (a/2, b/2))
        self._a = a/2
        self._b = b/2

    @property
    def a(self) -> float | int:
        """Side length along the x-axis."""
        return self._a

    @property
    def b(self) -> float | int:
        """Side length along the y-axis."""
        return self._b

    @property
    def size(self) -> np.ndarray:
        """2vector of side lengths (a, b)."""
        return np.asarray((self._a, self._b))


class RoundedRectangle(GenericGeometry):
    """
    Rounded rectangle defined by its side lengths and rounding radii for each corner.

    Args:
        a: Side length along the x-axis.
        b: Side length along the y-axis.
        rounding: rounding radii for each corner.
    """

    def __init__(self, a: float, b: float, rounding: tuple | list | np.ndarray):
        GenericGeometry.__init__(self, sdf_rounded_box_2d, (a/2, b/2), rounding[:4])
        self._a = a/2
        self._b = b/2
        self._round_corners = rounding[:4]

    @property
    def a(self) -> float | int:
        """Side length along the x-axis."""
        return self._a

    @property
    def b(self) -> float | int:
        """Side length along the y-axis."""
        return self._b

    @property
    def size(self) -> np.ndarray:
        """2vector of side lengths (a, b)."""
        return np.asarray((self._a, self._b))

    @property
    def round_corners(self) -> np.ndarray:
        """rounding on each of the corners."""
        return np.asarray(self._round_corners)


class Segment(GenericGeometry):
    """
    Line segment defined by its end points.

    Args:
        a: Vector defining the position of the start point.
        b: Vector defining the position of the end point.
    """

    def __init__(self, a: tuple | list | np.ndarray, b: tuple | list | np.ndarray):
        GenericGeometry.__init__(self, sdf_segment_2d, a, b)
        self._a = np.asarray(a)
        self._b = np.asarray(b)

    @property
    def point_a(self) -> np.ndarray:
        """Vector defining the position of the start point."""
        return self._a

    @property
    def point_b(self) -> np.ndarray:
        """Vector defining the position of the end point."""
        return self._b


class Triangle(GenericGeometry):
    """
    Triangle defined by the three vertices.

    Args:
        a: Vector defining the position of the first vertex.
        b: Vector defining the position of the second vertex.
        c: Vector defining the position of the third vertex.
    """

    def __init__(self, a: tuple | list | np.ndarray, b: tuple | list | np.ndarray, c: tuple | list | np.ndarray):
        GenericGeometry.__init__(self, sdf_triangle_2d, np.asarray(a), np.asarray(b), np.asarray(c))
        self._a = np.asarray(a)
        self._b = np.asarray(b)
        self._c = np.asarray(c)

    @property
    def a(self) -> np.ndarray:
        """Vector defining the position of the first vertex."""
        return self._a

    @property
    def b(self) -> np.ndarray:
        """Vector defining the position of the second vertex."""
        return self._b

    @property
    def c(self) -> np.ndarray:
        """Vector defining the position of the third vertex."""
        return self._c


class SectorOld(GenericGeometry):

    def __init__(self, radius, angle):
        GenericGeometry.__init__(self, sdf_sector_old, radius, angle/2)
        self._radius = radius
        self._angle = angle/2

    @property
    def radius(self):
        return self._radius

    @property
    def angle(self):
        return self._angle


class Sector(GenericGeometry):
    """
    Sector defined by the radius of the circle and two angles.

    Args:
        radius: Radius of the circle.
        angle_1: First angle defining the sector.
        angle_2: Second angle defining the sector.
    """

    def __init__(self, radius: float | int, angle_1: float | int, angle_2: float | int):
        GenericGeometry.__init__(self, sdf_sector, radius, angle_1, angle_2)
        self._radius = radius
        self._angle_1 = angle_1
        self._angle_2 = angle_2

    @property
    def radius(self) -> float | int:
        """Radius of the circle."""
        return self._radius

    @property
    def angle_1(self) -> float | int:
        """First angle defining the sector."""
        return self._angle_1

    @property
    def angle_2(self) -> float | int:
        """Second angle defining the sector."""
        return self._angle_2


class InfiniteSector(GenericGeometry):
    """
    Sector of infinite radius defined by two angles.

    Args:
        angle_1: First angle defining the sector.
        angle_2: Second angle defining the sector.
    """

    def __init__(self, angle_1: float | int, angle_2: float | int):
        GenericGeometry.__init__(self, sdf_inf_sector, angle_1, angle_2)
        self._angle_1 = angle_1
        self._angle_2 = angle_2

    @property
    def angle_1(self) -> float | int:
        """First angle defining the sector."""
        return self._angle_1

    @property
    def angle_2(self) -> float | int:
        """Second angle defining the sector."""
        return self._angle_2


class Arc(GenericGeometry):
    """
    Arc defined by the radius and the angles of both ends.

    Args:
        radius: Radius of the arc.
        start_angle: Angle of one end with respect to the x-axis.
        end_angle: Angle of the other end with respect to the x-axis.
    """

    def __init__(self, radius: float, start_angle: float, end_angle: float):
        GenericGeometry.__init__(self, sdf_arc, radius, start_angle, end_angle)
        self._radius = radius
        self._sangle = start_angle
        self._eangle = end_angle

    @property
    def radius(self):
        """Radius of the arc."""
        return self._radius

    @property
    def start_angle(self):
        """Angle of one end with respect to the x-axis."""
        return self._sangle

    @property
    def end_angle(self):
        """Angle of the other end with respect to the x-axis."""
        return self._eangle


class ParametricCurve(GenericGeometry):
    """
    Curve connecting points on the user provided parametric curve.

    Args:
        parametric_curve: Function defining the parametric curve with a single variable (t) and other parameters.
        parametric_curve_parameters: Parameters of the function defining the parametric curve.
        t_range: Range of the t parameter.
            t_range[0] -  start, t_range[1] - end, t_range[2] -  number of steps in between.
        closed: Is the curve closed on itself. False by default.
    """

    def __init__(self,
                 parametric_curve: Callable[[np.ndarray, tuple], np.ndarray],
                 parametric_curve_parameters: tuple,
                 t_range: tuple,
                 closed: bool = False):
        self._curve = parametric_curve
        self._c_params = parametric_curve_parameters
        self._t_range = t_range
        self._closed = closed

        GenericGeometry.__init__(self,
                                 self.sdf_closed_curve(),
                                 parametric_curve,
                                 parametric_curve_parameters,
                                 self.ts)

    def sdf_closed_curve(self) -> Callable[[np.ndarray, tuple], np.ndarray]:
        """
        Returns the SDF of the closed curve if parameter close==True.

        Returns:
            Modified SDF
        """
        def new_geo_object(co, f, f_parameters, t):

            p0 = f(t[0], *f_parameters)
            p1 = f(t[-1], *f_parameters)

            f1 = sdf_parametric_curve_2d(co, f, f_parameters, t)
            f2 = sdf_segment_2d(co, p0, p1)

            return np.minimum(f1, f2)

        if self.closed:
            return new_geo_object
        else:
            return sdf_parametric_curve_2d

    @property
    def steps(self) -> int:
        """Number of points at which the curve is evaluated."""
        return self._t_range[2]

    @property
    def t_start(self) -> float | int:
        """Start value of parameter t."""
        return self._t_range[0]

    @property
    def t_end(self) -> float | int:
        """End value of parameter t."""
        return self._t_range[1]

    @property
    def ts(self) -> np.ndarray:
        """Values of parameter t."""
        return np.linspace(*self._t_range)

    @property
    def closed(self) -> bool:
        """Is the segmented line closed on itself. False by default."""
        return self._closed

    def shape(self) -> Callable[[np.ndarray, tuple], np.ndarray]:
        """
        Converts the closed parametric curve into a shape.

        Returns:
            Modified SDF.
        """
        if not self.closed:
            return self.geo_object

        self._mod.append("shape")

        geo_object = self.geo_object

        def new_geo_object(co, *params):

            d = geo_object(co, *params)
            if np.any(d < 0):
                return geo_object(co, *params)

            ts_ = np.zeros(self.steps+1)
            ts_[:self.steps] = self.ts

            points = self._curve(ts_, *self._c_params)

            interior = np.ones(co.shape[1])
            for i in range(points.shape[1] - 1):
                t = points[:, i + 1] - points[:, i]
                t = t / np.linalg.norm(t)
                n = np.asarray([-t[1], t[0], 0])
                n[1] = n[1] - 2 * (n[1] < 0) * n[1]

                lx = np.minimum(points[0, i], points[0, i + 1])
                ux = np.maximum(points[0, i], points[0, i + 1])
                mask = (co[0, :] >= lx) * (co[0, :] < ux)

                s = np.sign(np.dot(co[:2, mask].T - points[:2, i], n[:2]))
                interior[mask] *= s

            return geo_object(co, *params)*interior

        self.geo_object = new_geo_object
        return new_geo_object


class SegmentedParametricCurve(GenericGeometry):
    """
    Segmented line connecting the user provided points.

    Args:
        points: Points to connect.
        t_range: Range of the t parameter.
            t_range[0] -  start, t_range[1] - end, t_range[2] -  number of steps in between.
            t_range[1]-t_range[0] >= number of points.
        closed: Is the segmented line closed on itself. False by default.
    """

    def __init__(self, points: list | tuple | np.ndarray, t_range: tuple, closed: bool = False):
        self._points = np.asarray(points)
        if self._points.shape[1] < self._points.shape[0]:
            self._points = self._points.T
        self._t_range = t_range
        self._closed = closed

        GenericGeometry.__init__(self,
                                 self.sdf_closed_curve(),
                                 self._points,
                                 self.ts)

    def sdf_closed_curve(self) -> Callable[[np.ndarray, tuple], np.ndarray]:
        """
        Returns the SDF of the closed curve if parameter close==True.

        Returns:
            Modified SDF
        """
        def new_geo_object(co, points, t):

            f1 = sdf_segmented_curve_2d(co, points, t)
            f2 = sdf_segment_2d(co, points[:, 0], points[:, -1])

            return np.minimum(f1, f2)

        if self.closed:
            return new_geo_object
        else:
            return sdf_segmented_curve_2d

    @property
    def steps(self) -> int:
        """Number of points at which the curve is evaluated."""
        return self._t_range[2]

    @property
    def t_start(self) -> float | int:
        """Start value of parameter t."""
        return self._t_range[0]

    @property
    def t_end(self) -> float | int:
        """End value of parameter t."""
        return self._t_range[1]

    @property
    def ts(self) -> np.ndarray:
        """Values of parameter t."""
        tt = np.linspace(self._t_range[0], self._t_range[1]-1, self._t_range[2])
        ts_ = np.clip(tt, 0, self._points.shape[1]-1.0001)
        return ts_

    @property
    def closed(self) -> bool:
        """Is the segmented line closed on itself. False by default."""
        return self._closed

    def polygon(self) -> Callable[[np.ndarray, tuple], np.ndarray]:
        """
        Converts the closed segmented line into a polygon.

        Returns:
            Modified SDF.
        """
        if not self.closed:
            return self.geo_object

        self._mod.append("polygon")

        geo_object = self.geo_object

        def new_geo_object(co, *params):

            d = geo_object(co, *params)
            if np.any(d < 0):
                return geo_object(co, *params)

            points = np.zeros((self._points.shape[0], self._points.shape[1] + 1))
            points[:, :self._points.shape[1]] = self._points
            points[:, -1] = self._points[:, 0]

            interior = np.ones(co.shape[1])
            for i in range(points.shape[1] - 1):
                t = points[:, i + 1] - points[:, i]
                t = t / np.linalg.norm(t)
                n = np.asarray([-t[1], t[0], 0])
                n[1] = n[1] - 2 * (n[1] < 0) * n[1]

                lx = np.minimum(points[0, i], points[0, i + 1])
                ux = np.maximum(points[0, i], points[0, i + 1])
                mask = (co[0, :] >= lx) * (co[0, :] < ux)

                s = np.sign(np.dot(co[:, mask].T - points[:, i], n))
                interior[mask] *= s

            return geo_object(co, *params) * interior

        self.geo_object = new_geo_object
        return new_geo_object


class SegmentedLine(GenericGeometry):
    """
    Segmented line connecting the user provided points.

    Args:
        points: Points to connect.
        closed: Is the segmented line closed on itself. False by default.
    """

    def __init__(self, points: list | tuple | np.ndarray, closed: bool = False):
        self._points = np.asarray(points)
        if self._points.shape[1] < self._points.shape[0]:
            self._points = self._points.T
        self._closed = closed

        GenericGeometry.__init__(self,
                                 self.sdf_closed_curve(),
                                 self._points)

    def sdf_closed_curve(self) -> Callable[[np.ndarray, tuple], np.ndarray]:
        """
        Returns the SDF of the closed curve if parameter close==True.

        Returns:
            Modified SDF
        """
        def new_geo_object(co, points):

            f1 = sdf_segmented_line_2d(co, points)
            f2 = sdf_segment_2d(co, points[:, 0], points[:, -1])

            return np.minimum(f1, f2)

        if self.closed:
            return new_geo_object
        else:
            return sdf_segmented_line_2d

    @property
    def closed(self) -> bool:
        """Is the segmented line closed on itself. False by default."""
        return self._closed

    def polygon(self) -> Callable[[np.ndarray, tuple], np.ndarray]:
        """
        Converts the closed segmented line into a polygon.

        Returns:
            Modified SDF.
        """
        if not self.closed:
            return self.geo_object

        self._mod.append("polygon")

        geo_object = self.geo_object

        def new_geo_object(co, *params):

            d = geo_object(co, *params)
            if np.any(d<0):
                return geo_object(co, *params)

            points = np.zeros((self._points.shape[0], self._points.shape[1] + 1))
            points[:, :self._points.shape[1]] = self._points
            points[:, -1] = self._points[:, 0]

            interior = np.ones(co.shape[1])
            for i in range(points.shape[1]-1):
                t = points[:, i+1] - points[:, i]
                t = t/np.linalg.norm(t)
                n = np.asarray([-t[1], t[0], 0])
                n[1] = n[1] - 2*(n[1] < 0)*n[1]

                lx = np.minimum(points[0, i], points[0, i+1])
                ux = np.maximum(points[0, i], points[0, i + 1])
                mask = (co[0, :] >= lx) * (co[0, :] < ux)

                s = np.sign(np.dot(co[:, mask].T - points[:, i], n))
                interior[mask] *= s

            return geo_object(co, *params)*interior

        self.geo_object = new_geo_object
        return new_geo_object


class PointCloud2D(GenericGeometry):
    """
    SDF of the point cloud.

    Args:
        points: Positions of the points in an array of shape (2, N-points).
    """

    def __init__(self, points: list | tuple | np.ndarray):
        self._points = np.asarray(points)
        if self._points.shape[1] < self._points.shape[0]:
            self._points = self._points.T

        GenericGeometry.__init__(self, sdf_point_cloud_2d, self._points)

    @property
    def points(self) -> np.ndarray:
        """Positions of the points in an array of shape (2, N-points)."""
        return self._points



# Copyright (C) 2024 Peter Ropač
# This file is part of SPOMSO.
# SPOMSO is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
# SPOMSO is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.
# You should have received a copy of the GNU Lesser General Public License along with SPOMSO. If not, see <https://www.gnu.org/licenses/>.

import numpy as np
from typing import Callable
from spomso.cores.transformations import EuclideanTransform
from spomso.cores.modifications import ModifyObject
from spomso.cores.geom import GenericGeometry
from spomso.cores.sdf_2D import sdf_circle
from spomso.cores.sdf_3D import sdf_sphere, sdf_cylinder, sdf_box, sdf_torus, sdf_arc_3d, sdf_chainlink, sdf_braid
from spomso.cores.sdf_3D import sudf_plane, sdf_plane, sdf_segment_3d, sdf_cone
from spomso.cores.sdf_3D import sdf_infinite_cone, sdf_oriented_infinite_cone
from spomso.cores.sdf_3D import sdf_solid_angle, sdf_triangle_3d, sdf_quad_3d
from spomso.cores.sdf_3D import sdf_parametric_curve_3d, sdf_segmented_curve_3d, sdf_segmented_line_3d
from spomso.cores.sdf_3D import sdf_x, sdf_y, sdf_z
from spomso.cores.sdf_3D import sdf_point_cloud_3d


class GenericGeometry3D(EuclideanTransform, ModifyObject):

    def __init__(self, geo_sdf, *geo_parameters):
        EuclideanTransform.__init__(self)
        ModifyObject.__init__(self, geo_sdf)
        self.geo_parameters = geo_parameters
        self._sdf = geo_sdf

    def create(self, co):
        self._sdf = self.modified_object
        return self.apply(self._sdf, co, self.geo_parameters)

    def propagate(self, co, *parameters_):
        self._sdf = self.modified_object
        return self.apply(self._sdf, co, self.geo_parameters)


class X(GenericGeometry):
    """
    Value of the X coordinate zeroed at some offset value.

    Args:
        offset: Value at which the field is zero.
    """

    def __init__(self, offset: float | int):
        GenericGeometry.__init__(self, sdf_x, offset)
        self._offset = offset

    @property
    def offset(self) -> float | int:
        """Value at which the field is zero."""
        return self._offset


class Y(GenericGeometry):
    """
    Value of the Y coordinate zeroed at some offset value.

    Args:
        offset: Value at which the field is zero.
    """

    def __init__(self, offset: float | int):
        GenericGeometry.__init__(self, sdf_y, offset)
        self._offset = offset

    @property
    def offset(self) -> float | int:
        """Value at which the field is zero."""
        return self._offset


class Z(GenericGeometry):
    """
    Value of the Z coordinate zeroed at some offset value.

    Args:
        offset: Value at which the field is zero.
    """

    def __init__(self, offset: float | int):
        GenericGeometry.__init__(self, sdf_z, offset)
        self._offset = offset

    @property
    def offset(self) -> float | int:
        """Value at which the field is zero."""
        return self._offset


class InfiniteCylinder(GenericGeometry):
    """
    Cylinder of a given radius and infinite height.

    Args:
        radius: Radius of the cylinder.
    """

    def __init__(self, radius: float | int):
        GenericGeometry.__init__(self, sdf_circle, radius)
        self._radius = radius

    @property
    def radius(self) -> float | int:
        """Radius of the cylinder."""
        return self._radius


class Cylinder(GenericGeometry):
    """
    Cylinder defined by the radius and height.

    Args:
        radius: Radius of the cylinder.
        height: Height of the cylinder.
    """

    def __init__(self, radius: float | int, height: float | int):
        GenericGeometry.__init__(self, sdf_cylinder, radius, height)
        self._radius = radius
        self._height = height

    @property
    def radius(self) -> float | int:
        """Radius of the cylinder."""
        return self._radius

    @property
    def height(self) -> float | int:
        """Radius of the cylinder."""
        return self._height


class Sphere(GenericGeometry):
    """
    Sphere defined by the radius.

    Args:
        radius: Radius of the sphere.
    """

    def __init__(self, radius: float):
        GenericGeometry.__init__(self, sdf_sphere, radius)
        self._radius = radius

    @property
    def radius(self) -> float | int:
        """Radius of the sphere."""
        return self._radius


class Box(GenericGeometry):
    """
    Box defined by the side lengths a, b, c.

    Args:
        a: Side length along the x-axis.
        b: Side length along the y-axis.
        c: Side length along the z-axis.
    """
    
    def __init__(self, a: float | int, b: float | int, c: float | int):
        GenericGeometry.__init__(self, sdf_box, a/2, b/2, c/2)
        self._a = a/2
        self._b = b/2
        self._c = c/2

    @property
    def a(self) -> float | int:
        """Side length along the x-axis."""
        return self._a

    @property
    def b(self) -> float | int:
        """Side length along the y-axis."""
        return self._b

    @property
    def c(self) -> float | int:
        """Side length along the z-axis."""
        return self._c


class Plane(GenericGeometry):
    """
    Plane defined by the normal vector.
    
    Args:
        normal: Normal vector of the plane.
        thickness: Thickness of the plane.
    """
    
    def __init__(self, normal: tuple | list | np.ndarray, thickness: float | int):
        GenericGeometry.__init__(self, sudf_plane, np.asarray(normal), thickness)
        self._normal = np.asarray(normal)
        self._thickness = thickness

    @property
    def normal(self) -> np.ndarray:
        """Normal vector of the plane."""
        return self._normal

    @property
    def thickness(self) -> float | int:
        """Thickness of the plane."""
        return self._thickness


class OrientedPlane(GenericGeometry):
    """
    Plane defined by a normal vector. SDF has a negative value for points below the plane.
    Args:
        normal: Normal vector of the plane.
        offset: Offset of the origin along the normal vector.
    """
    
    def __init__(self, normal: tuple | list | np.ndarray, offset: int | float):
        GenericGeometry.__init__(self, sdf_plane, np.asarray(normal), offset)
        self._normal = np.asarray(normal)
        self._offset = offset

    @property
    def normal(self) -> np.ndarray:
        """Normal vector of the plane."""
        return self._normal

    @property
    def offset(self) -> float | int:
        """Offset of the origin along the normal vector."""
        return self._offset


class Line(GenericGeometry):
    """
    Line defined by the starting and ending point.
    
    Args:
        a: Vector defining the starting point.
        b: Vector defining the ending point.
    """

    def __init__(self, a: np.ndarray | list | tuple, b: np.ndarray | list | tuple):
        GenericGeometry.__init__(self, sdf_segment_3d, a, b)
        self._a = np.asarray(a)
        self._b = np.asarray(b)

    @property
    def point_a(self) -> np.ndarray:
        """Vector defining the starting point."""
        return self._a

    @property
    def point_b(self) -> np.ndarray:
        """Vector defining the ending point."""
        return self._b
    
    
class Triangle3D(GenericGeometry):
    """
    Triangle defined by the three vertices.

    Args:
        a: Vector defining the position of the first vertex.
        b: Vector defining the position of the second vertex.
        c: Vector defining the position of the third vertex.
    """

    def __init__(self, a: tuple | list | np.ndarray, b: tuple | list | np.ndarray, c: tuple | list | np.ndarray):
        GenericGeometry.__init__(self, sdf_triangle_3d, np.asarray(a), np.asarray(b), np.asarray(c))
        self._a = np.asarray(a)
        self._b = np.asarray(b)
        self._c = np.asarray(c)

    @property
    def a(self) -> np.ndarray:
        """Vector defining the position of the first vertex."""
        return self._a

    @property
    def b(self) -> np.ndarray:
        """Vector defining the position of the second vertex."""
        return self._b

    @property
    def c(self) -> np.ndarray:
        """Vector defining the position of the third vertex."""
        return self._c


class Quad(GenericGeometry):
    """
    Quadrilateral defined by the four vertices.
    
    Args:
        a: Vector defining the position of the first vertex.
        b: Vector defining the position of the second vertex.
        c: Vector defining the position of the third vertex.
        d: Vector defining the position of the fourth vertex.
    """
    def __init__(self,
                a: tuple | list | np.ndarray, b: tuple | list | np.ndarray,
                c: tuple | list | np.ndarray, d: tuple | list | np.ndarray):
        GenericGeometry.__init__(self, sdf_quad_3d, np.asarray(a), np.asarray(b), np.asarray(c), np.asarray(d))
        self._a = np.asarray(a)
        self._b = np.asarray(b)
        self._c = np.asarray(c)
        self._d = np.asarray(d)
        
    @property
    def a(self) -> np.ndarray:
        """Vector defining the position of the first vertex."""
        return self._a

    @property
    def b(self) -> np.ndarray:
        """Vector defining the position of the second vertex."""
        return self._b

    @property
    def c(self) -> np.ndarray:
        """Vector defining the position of the third vertex."""
        return self._c

    @property
    def d(self) -> np.ndarray:
        """Vector defining the position of the fourth vertex."""
        return self._d


class Torus(GenericGeometry):
    """
    Torus defined by the primary and the secondary radius.

    Args:
        primary_radius: Primary radius of the torus.
        secondary_radius: Secondary radius of the torus.
    """

    def __init__(self, primary_radius: float, secondary_radius: float):
        GenericGeometry.__init__(self, sdf_torus, primary_radius, secondary_radius)
        self._pradius = primary_radius
        self._sradius = secondary_radius

    @property
    def primary_radius(self) -> float | int:
        """Primary radius of the torus - distance between the center of the tube and the center of the torus."""
        return self._pradius

    @property
    def secondary_radius(self) -> float | int:
        """Secondary radius of the torus - radius of the tube."""
        return self._sradius


class ChainLink(GenericGeometry):
    """
    Chain Link defined by the primary radius, secondary radius and the length.

    Args:
        primary_radius: Width of the chain link.
        secondary_radius: Thickness of the wire.
        length: Length of the chain link.
    """

    def __init__(self, primary_radius: float | int, secondary_radius: float | int, length: float | int):
        GenericGeometry.__init__(self, sdf_chainlink, primary_radius, secondary_radius, length/2)
        self._pradius = primary_radius
        self._sradius = secondary_radius
        self._length = length/2

    @property
    def primary_radius(self) -> float | int:
        """Width of the chain link."""
        return self._pradius

    @property
    def secondary_radius(self) -> float | int:
        """Thickness of the wire."""
        return self._sradius

    @property
    def length(self) -> float | int:
        """Length of the chain link."""
        return self._length


class Braid(GenericGeometry):
    """
    Braid defined as a twisted chain link, where the total number of revolutions is defined via the pitch.

    Args:
        length: Length of the braid.
        primary_radius: Width of the braid.
        secondary_radius: Thickness of the braid.
        pitch: Pitch of the twist in units rad/unit length.
    """

    def __init__(self, length: float | int,
                primary_radius: float | int, secondary_radius: float | int,
                pitch: float | int):
        GenericGeometry.__init__(self, sdf_braid, length/2, primary_radius, secondary_radius, pitch)
        self._pradius = primary_radius
        self._sradius = secondary_radius
        self._length = length/2
        self._pitch = pitch

    @property
    def primary_radius(self) -> float | int:
        """Width of the braid."""
        return self._pradius

    @property
    def secondary_radius(self) -> float | int:
        """Thickness of the braid."""
        return self._sradius

    @property
    def length(self) -> float | int:
        """Length of the braid."""
        return self._length

    @property
    def pitch(self) -> float | int:
        """Pitch of the twist in units rad/unit length."""
        return self._pitch


class Arc3D(GenericGeometry):
    """
    Arc defined by the radius, thickness, and the angles of both ends.
        
    Args:    
        radius: Radius of the arc.
        thickness: Thickness of the arc.
        start_angle: Angle of one end with respect to the x-axis.
        end_angle: Angle of the other end with respect to the x-axis.
    """

    def __init__(self,
                radius: float | int, thickness: float | int,
                start_angle: float | int, end_angle: float | int):
        GenericGeometry.__init__(self, sdf_arc_3d,
                                radius, thickness,
                                start_angle, end_angle)
        self._pradius = radius
        self._sradius = thickness
        self._sangle = start_angle
        self._eangle = end_angle

    @property
    def start_angle(self) -> float | int:
        """Angle of one end with respect to the x-axis."""
        return self._sangle

    @property
    def end_angle(self) -> float | int:
        """Angle of the other end with respect to the x-axis."""
        return self._eangle

    @property
    def primary_radius(self) -> float | int:
        """Radius of the arc."""
        return self._pradius

    @property
    def secondary_radius(self) -> float | int:
        """Thickness of the arc."""
        return self._sradius


class Cone(GenericGeometry):

    def __init__(self, height: float | int, angle: float | int):
        """
        Cone defined by its height and the angle of the slope.
        The base of the cone is moved down by: height - height_offset.

        Args:
            height: Height of the cone.
            angle: Angle of the slope.
        """
        GenericGeometry.__init__(self, sdf_cone, height, angle)
        self._height = height
        self._angle = angle

    @property
    def height(self) -> float | int:
        """Height of the cone."""
        return self._height

    @property
    def height_offset(self) -> float | int:
        """Center of mass height."""
        return self._height*(0.5**(1/3))

    @property
    def angle(self) -> float | int:
        """Angle of the slope."""
        return self._angle

    @property
    def base_radius(self) -> float | int:
        """Radius of the base."""
        return self._height*np.tan(self._angle)


class InfiniteCone(GenericGeometry):
    """
    Cone with infinite height defined by the angle of its slope. The tip of the cone is in the origin.

    Args:
        angle: Angle of the slope.
    """

    def __init__(self, angle: float | int):
        GenericGeometry.__init__(self, sdf_infinite_cone, angle)
        self._angle = angle

    @property
    def angle(self) -> float | int:
        """Angle of the slope."""
        return self._angle


class OrientedInfiniteCone(GenericGeometry):
    """
    Cone with infinite height defined by the angle of its slope. The tip of the cone is in the origin.
    Values of the SDF below the cone are negative.

    Args:
        angle: Angle of the slope.
    """
    def __init__(self, angle: float | int):
        GenericGeometry.__init__(self, sdf_oriented_infinite_cone, angle)
        self._angle = angle

    @property
    def angle(self) -> float | int:
        """Angle of the slope."""
        return self._angle


class SolidAngle(GenericGeometry):
    """
    Solid angle defined by the radius of the globe and two angles.

    Args:
        radius: Radius of the globe.
        angle_1: First angle defining the sector.
        angle_2: Second angle defining the sector.
    """

    def __init__(self, radius: float | int, angle_1: float | int, angle_2: float | int):
        GenericGeometry.__init__(self, sdf_solid_angle, radius, angle_1, angle_2)
        self._radius = radius
        self._angle_1 = angle_1
        self._angle_2 = angle_2

    @property
    def radius(self) -> float | int:
        """Radius of the globe."""
        return self._radius

    @property
    def angle_1(self) -> float | int:
        """First angle defining the sector."""
        return self._angle_1

    @property
    def angle_2(self) -> float | int:
        """Second angle defining the sector."""
        return self._angle_2


class ParametricCurve3D(GenericGeometry):
    """
    Curve connecting the points on the user provided parametric curve.

    Args:
        parametric_curve: Function defining the parametric curve with a single variable (t) and other parameters.
        parametric_curve_parameters: Parameters of the function defining the parametric curve.
        t_range: Range of the t parameter/variable.
            t_range[0] -  start, t_range[1] - end, t_range[2] -  number of steps in between.
        closed: Is the curve closed on itself. False by default.
    """

    def __init__(self,
                 parametric_curve: Callable[[np.ndarray, tuple], np.ndarray],
                 parametric_curve_parameters: tuple,
                 t_range: tuple,
                 closed: bool = False):

        self._curve = parametric_curve
        self._c_params = parametric_curve_parameters
        self._t_range = t_range
        self._closed = closed

        GenericGeometry.__init__(self, self.sdf_closed_curve(),
                                 parametric_curve, parametric_curve_parameters, self.ts)

    def sdf_closed_curve(self) -> Callable[[np.ndarray, tuple], np.ndarray]:
        """
        Returns the SDF of the closed curve if parameter close==True.

        Returns:
            Modified SDF
        """
        def new_geo_object(co, f, f_parameters, t):

            p0 = f(t[0], *f_parameters)
            p1 = f(t[-1], *f_parameters)

            f1 = sdf_parametric_curve_3d(co, f, f_parameters, t)
            f2 = sdf_segment_3d(co, p0, p1)

            return np.minimum(f1, f2)

        if self.closed:
            return new_geo_object
        else:
            return sdf_parametric_curve_3d

    @property
    def steps(self) -> float | int:
        """Number of points at which the curve is evaluated."""
        return self._t_range[2]

    @property
    def t_start(self) -> float | int:
        """Start value of parameter t."""
        return self._t_range[0]

    @property
    def t_end(self) -> float | int:
        """End value of parameter t."""
        return self._t_range[1]

    @property
    def ts(self) -> np.ndarray:
        """Values of parameter t."""
        return np.linspace(*self._t_range)

    @property
    def closed(self) -> bool:
        """Is the segmented line closed on itself. False by default."""
        return self._closed


class SegmentedParametricCurve3D(GenericGeometry):
    """
    Segmented line connecting the user provided points.
        points: Points to connect.
        t_range: Range of the t parameter.
            t_range[0] -  start, t_range[1] - end, t_range[2] -  number of steps in between.
            t_range[1]-t_range[0] >= number of points.
        closed: Is the segmented line closed on itself. False by default.
    """

    def __init__(self, points: list | tuple | np.ndarray, t_range: tuple, closed: bool = False):
        self._points = np.asarray(points)
        if self._points.shape[1] < self._points.shape[0]:
            self._points = self._points.T
        self._t_range = t_range
        self._closed = closed

        GenericGeometry.__init__(self, self.sdf_closed_curve(), self._points, self.ts)

    def sdf_closed_curve(self) -> Callable[[np.ndarray, tuple], np.ndarray]:
        """
        Returns the SDF of the closed curve if parameter close==True.

        Returns:
            Modified SDF
        """
        def new_geo_object(co, points, t):

           f1 = sdf_segmented_curve_3d(co, points, t)
           f2 = sdf_segment_3d(co, points[:, 0], points[:, -1])

           return np.minimum(f1, f2)

        if self.closed:
           return new_geo_object
        else:
           return sdf_segmented_curve_3d

    @property
    def steps(self) -> int:
        """Number of points at which the curve is evaluated."""
        return self._t_range[2]

    @property
    def t_start(self) -> float | int:
        """Start value of parameter t."""
        return self._t_range[0]

    @property
    def t_end(self) -> float | int:
        """End value of parameter t."""
        return self._t_range[1]

    @property
    def ts(self) -> np.ndarray:
        """Values of parameter t."""
        tt = np.linspace(self._t_range[0], self._t_range[1]-1, self._t_range[2]) - self._t_range[0]
        ts_ = np.clip(tt, 0, self._points.shape[1]-1.0001)
        return ts_

    @property
    def closed(self) -> bool:
        """Is the segmented line closed on itself. False by default."""
        return self._closed


class SegmentedLine3D(GenericGeometry):
    """
    Segmented line connecting the user provided points.

    Args:
        points: Points to connect.
        closed: Is the segmented line closed on itself. False by default.
    """

    def __init__(self, points: list | tuple | np.ndarray, closed: bool = False):
        self._points = np.asarray(points)
        if self._points.shape[1] < self._points.shape[0]:
            self._points = self._points.T
        self._closed = closed

        GenericGeometry.__init__(self, self.sdf_closed_curve(), self._points)

    def sdf_closed_curve(self) -> Callable[[np.ndarray, tuple], np.ndarray]:
        """
        Returns the SDF of the closed curve if parameter close==True.

        Returns:
            Modified SDF
        """
        def new_geo_object(co, points):

            f1 = sdf_segmented_line_3d(co, points)
            f2 = sdf_segment_3d(co, points[:, 0], points[:, -1])

            return np.minimum(f1, f2)

        if self.closed:
            return new_geo_object
        else:
            return sdf_segmented_curve_3d

    @property
    def closed(self) -> bool:
        """Is the segmented line closed on itself. False by default."""
        return self._closed


class PointCloud3D(GenericGeometry):
    """
    SDF of the point cloud.

    Args:
        points: Positions of the points in an array of shape (3, N-points).
    """

    def __init__(self, points: list | tuple | np.ndarray):
        self._points = np.asarray(points)
        if self._points.shape[1] < self._points.shape[0]:
            self._points = self._points.T

        GenericGeometry.__init__(self, sdf_point_cloud_3d, self._points)

    @property
    def points(self) -> np.ndarray:
        """Positions of the points in an array of shape (2, N-points)."""
        return self._points





from setuptools import setup, find_packages

VERSION = '1.2.1'
DESCRIPTION = 'SPOMSO'
LONG_DESCRIPTION = 'Python package for generating geometry with SDFs.'

# Setting up
setup(
    name="SPOMSO",
    version=VERSION,
    author="Peter Ropač",
    author_email="<peter.ropac@fmf.uni-lj.si>",
    description=DESCRIPTION,
    long_description=LONG_DESCRIPTION,
    url='https://github.com/peterropac/Aegolius/tree/main/Code',
    packages=find_packages(),
    install_requires=['numpy', 'scipy'],
    keywords=['python', 'geometry'],
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Education",
        "Programming Language :: Python :: 3",
        "Operating System :: OS Independent",
    ]
)


# Copyright (C) 2024 Peter Ropač
# This file is part of SPOMSO.
# SPOMSO is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
# SPOMSO is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.
# You should have received a copy of the GNU Lesser General Public License along with SPOMSO. If not, see <https://www.gnu.org/licenses/>.

import numpy as np
from scipy.ndimage import convolve
from typing import Callable
from spomso.cores.helper_functions import smarter_reshape


class PostProcess:
    """Class containing all the possible post-processing operations which can be applied to a scalar field.

    Attributes:
        unprocessed_geo_object: Original scalar field.
        processed_geo_object: Modified scalar field.

    Args:
        geo_object: Scalar field.
    """

    def __init__(self, geo_object: Callable[[np.ndarray, tuple], np.ndarray]):
        self._pmod = []
        self.unprocessed_geo_object = geo_object
        self.processed_geo_object = geo_object

    @property
    def post_processing_operations(self) -> list:
        """
        All the post-processing operations which were applied to the geometry in chronological order.
        
        Returns:
             List of post-processing operations.
        """
        return self._pmod

    @property
    def processed_object(self) -> Callable[[np.ndarray, tuple], np.ndarray]:
        """
        SDF of the modified geometry.
        
        Returns:
             SDF of the modified geometry.
        """
        return self.processed_geo_object

    @property
    def unprocessed_object(self) -> Callable[[np.ndarray, tuple], np.ndarray]:
        """
        SDF of the unmodified geometry.
        
        Returns:
             SDF of the unmodified geometry.
        """
        return self.unprocessed_geo_object

    def sigmoid_falloff(self, amplitude: float | int, width: float | int) -> Callable[[np.ndarray, tuple], np.ndarray]:
        """
        Applies a sigmoid to the scalar (Signed Distance Function) field.
        
        Args:
            amplitude: Maximum value of the transformed scalar field.
            width: Width of the sigmoid.
        
        Returns:
             Modified Scalar Field.
        """
        
        self._pmod.append("sigmoid_falloff")
        geo_object = self.processed_geo_object

        def new_geo_object(co, *params):
            u = geo_object(co, *params)
            return sigmoid_falloff(u, amplitude, width)

        self.processed_geo_object = new_geo_object
        return new_geo_object

    def positive_sigmoid_falloff(self,
                                 amplitude: float | int,
                                 width: float | int) -> Callable[[np.ndarray, tuple], np.ndarray]:
        """
        Applies a sigmoid, shifted to the positive velues by the value of the width parameter,
        to the scalar (Signed Distance Function) field.
        
        Args:
            amplitude: Maximum value of the transformed scalar field.
            width: Width of the sigmoid.
        
        Returns:
             Modified Scalar Field.
        """
        
        self._pmod.append("positive_sigmoid_falloff")
        geo_object = self.processed_geo_object

        def new_geo_object(co, *params):
            u = geo_object(co, *params)
            return positive_sigmoid_falloff(u, amplitude, width)

        self.processed_geo_object = new_geo_object
        return new_geo_object

    def capped_exponential(self,
                           amplitude: float | int,
                           width: float | int) -> Callable[[np.ndarray, tuple], np.ndarray]:
        """
        Applies a decreasing exponential functon to the scalar (Signed Distance Function) field.
        to the scalar (Signed Distance Function) field.
        
        Args:
            amplitude: Maximum value of the transformed scalar field.
            width: Range at which the value of the transformed scalar field drops to almost zero.
        
        Returns:
             Modified Scalar Field.
        """
        
        self._pmod.append("capped_exponential")
        geo_object = self.processed_geo_object

        def new_geo_object(co, *params):
            u = geo_object(co, *params)
            return capped_exponential(u, amplitude, width)

        self.processed_geo_object = new_geo_object
        return new_geo_object

    def hard_binarization(self, threshold: float) -> Callable[[np.ndarray, tuple], np.ndarray]:
        """
        Binarizes the Signed Distance field/pattern based on a threshold.
        Values below the threshold are 1 and values above are 0.
        
        Args:
            threshold: Binarization threshold.
        
        Returns:
             Modified Scalar Field.
        """
        
        self._pmod.append("hard_binarization")
        geo_object = self.processed_geo_object

        def new_geo_object(co, *params):
            u = geo_object(co, *params)
            return hard_binarization(u, threshold)

        self.processed_geo_object = new_geo_object
        return new_geo_object

    def linear_falloff(self, amplitude: float | int, width: float | int) -> Callable[[np.ndarray, tuple], np.ndarray]:
        """
        Applies a decreasing linear function to the scalar (Signed Distance Function) field.
        
        Args:
            amplitude: Maximum value of the transformed scalar field.
            width: Range at which the value of the transformed scalar field drops to zero.
        
        Returns:
             Modified Scalar Field.
        """
        
        self._pmod.append("linear_falloff")
        geo_object = self.processed_geo_object

        def new_geo_object(co, *params):
            u = geo_object(co, *params)
            return linear_falloff(u, amplitude, width)

        self.processed_geo_object = new_geo_object
        return new_geo_object

    def relu(self, width: float | int) -> Callable[[np.ndarray, tuple], np.ndarray]:
        """
        Applies the ReLU function to the scalar (Signed Distance Function) field.
        
        Args:
            width: Range at which the value of the transformed field reaches one.
        
        Returns:
             Modified Scalar Field.
        """
        
        self._pmod.append("relu")
        geo_object = self.processed_geo_object

        def new_geo_object(co, *params):
            u = geo_object(co, *params)
            return relu(u, width)

        self.processed_geo_object = new_geo_object
        return new_geo_object

    def smooth_relu(self, smooth_width: float | int,
                    width: float | int = 1, threshold: int | float = 0.01) -> Callable[[np.ndarray, tuple], np.ndarray]:
        """
        Applies the "squareplus" function to the scalar (Signed Distance Function) field.
        https://en.wikipedia.org/wiki/Rectifier_(neural_networks)
        
        Args:
            smooth_width: Distance from the origin at which the Smooth ReLU function
                is greater than ReLU for less than the value of the threshold parameter.
            width: Range at which the value of the transformed field reaches one.
            threshold: At smooth_width distance from the origin the value of the Smooth ReLU function is greater
                than ReLU for the value of the threshold parameter.
                at smooth_width distance from the origin.
        
        Returns:
            Transformed scalar field.
        """
        
        self._pmod.append("smooth_relu")
        geo_object = self.processed_geo_object

        def new_geo_object(co, *params):
            u = geo_object(co, *params)
            return smooth_relu(u, smooth_width, width, threshold)

        self.processed_geo_object = new_geo_object
        return new_geo_object

    def slowstart(self, smooth_width: float | int,
                  width: float | int = 1,
                  threshold: int | float = 0.01,
                  ground: bool = True) -> Callable[[np.ndarray, tuple], np.ndarray]:
        """
        Applies the SlowStart function to the scalar (Signed Distance Function) field.
        
        Args:
            smooth_width: Distance from the origin at which the SlowStart function
                is greater than ReLU for less than the value of the threshold parameter.
            width: Range at which the value of the transformed field reaches one.
            threshold: At smooth_width distance from the origin the value of the SlowStart function is greater
                than ReLU for the value of the threshold parameter.
            ground: if True the value of the function is zero at zero.
        
        Returns:
            Transformed scalar field.
        """

        self._pmod.append("slowstart")
        geo_object = self.processed_geo_object

        def new_geo_object(co, *params):
            u = geo_object(co, *params)
            return slowstart(u, smooth_width, width, threshold, ground=ground)

        self.processed_geo_object = new_geo_object
        return new_geo_object

    def gaussian_boundary(self, amplitude: float | int, width: float | int) -> Callable[[np.ndarray, tuple], np.ndarray]:
        """
        Applies the Gaussian to the scalar (Signed Distance Function) field.
        
        Args:
            amplitude: Maximum value of the transformed scalar field.
            width: Range at which the value of the transformed scalar field drops to almost zero.
        
        Returns:
             Modified Scalar Field.
        """
        
        self._pmod.append("gaussian_boundary")
        geo_object = self.processed_geo_object

        def new_geo_object(co, *params):
            u = geo_object(co, *params)
            return gaussian_boundary(u, amplitude, width)

        self.processed_geo_object = new_geo_object
        return new_geo_object

    def gaussian_falloff(self, amplitude: float | int, width: float | int) -> Callable[[np.ndarray, tuple], np.ndarray]:
        """
        Applies the Gaussian to the positive values of the scalar (Signed Distance Function) field.
        
        Args:
            amplitude: Maximum value of the transformed scalar field (and points at which the scalar field was < 0).
            width: Range at which the value of the transformed scalar field drops to almost zero.
        
        Returns:
             Modified Scalar Field.
        """
        
        self._pmod.append("gaussian_falloff")
        geo_object = self.processed_geo_object

        def new_geo_object(co, *params):
            u = geo_object(co, *params)
            return gaussian_falloff(u, amplitude, width)

        self.processed_geo_object = new_geo_object
        return new_geo_object

    def conv_averaging(self,
                       kernel_size: int | tuple | list | np.ndarray,
                       iterations: int,
                       co_resolution: tuple) -> Callable[[np.ndarray, tuple], np.ndarray]:
        """
        Averages the field using an averaging convolutional kernel of the specified size.
        
        Args:
            kernel_size: Size of the averaging kernel. Must be an integer or a tuple/array of the
                same dimension as the scaler field.
            iterations: Number of times the convolutional averaging is applied to the input scalar field.
            co_resolution: Resolution of the coordinate system.
        
        Returns:
             Modified Scalar Field.
        """
        
        self._pmod.append("conv_averaging")
        geo_object = self.processed_geo_object

        def new_geo_object(co, *params):
            u = geo_object(co, *params)
            u = smarter_reshape(u, co_resolution)
            out = conv_averaging(u, kernel_size, iterations)
            return out.flatten()

        self.processed_geo_object = new_geo_object
        return new_geo_object

    def conv_edge_detection(self,
                            co_resolution: tuple) -> Callable[[np.ndarray, tuple], np.ndarray]:
        """
        Edge detection based on a 3x3 convolutional kernel.
        
        Args:
            co_resolution: Resolution of the coordinate system.
        
        Returns:
             Modified Scalar Field.
        """
        
        self._pmod.append("conv_edge_detection")
        geo_object = self.processed_geo_object

        def new_geo_object(co, *params):
            u = geo_object(co, *params)
            u = smarter_reshape(u, co_resolution)
            return conv_edge_detection(u)

        self.processed_geo_object = new_geo_object
        return new_geo_object

    def custom_post_process(self,
                            function: Callable[[np.ndarray, tuple], np.ndarray],
                            parameters: tuple,
                            post_process_name: str = "custom") -> Callable[[np.ndarray, tuple], np.ndarray]:
        """
        Applies a custom user-specified post-processing function to a scalar (Signed Distance Function) field.
        
        Args:
            function: A custom post-processing function which takes the SDF as a first argument
                and the parameters of the function as the next arguments.
            parameters: Parameters of the custom post-processing function.
            post_process_name: Name of the custom post-processing function.
        
        Returns:
             Modified Scalar Field.
        """
        
        self._pmod.append(post_process_name)
        geo_object = self.processed_geo_object

        def new_geo_object(co, *params):
            u = geo_object(co, *params)
            return function(u, *parameters)

        self.processed_geo_object = new_geo_object
        return new_geo_object


# ----------------------------------------------------------------------------------------------------------------------
# SCALAR POSTPROCESSING FUNCTIONS

def sigmoid_falloff(u: np.ndarray, amplitude: float | int, width: float | int) -> np.ndarray:
    """
    Applies a sigmoid to the scalar (Signed Distance Function) field.
    
    Args:
        u: Signed Distance field or any scalar field.
        amplitude: Maximum value of the transformed scalar field.
        width: Width of the sigmoid.
    
    Returns:
        Transformed scalar field.
    """
    e = np.exp(4 * u / width)

    return amplitude*(1/(1 + e))


def positive_sigmoid_falloff(u: np.ndarray, amplitude: float | int, width: float | int) -> np.ndarray:
    """
    Applies a sigmoid, shifted to the positive velues by the value of the width parameter,
    to the scalar (Signed Distance Function) field.
    
    Args:
        u: Signed Distance field or any scalar field.
        amplitude: Maximum value of the transformed scalar field.
        width: Width of the sigmoid.
    
    Returns:
        Transformed scalar field.
    """
    e = np.exp(4 * (u - width) / width)

    return amplitude*(1/(1 + e))


def capped_exponential(u: np.ndarray, amplitude: float | int, width: float | int) -> np.ndarray:
    """
    Applies a decreasing exponential functon to the scalar (Signed Distance Function) field.
    
    Args:
        u: Signed Distance field or any scalar field.
        amplitude: Maximum value of the transformed scalar field.
        width: Range at which the value of the transformed scalar field drops to almost zero.
    
    Returns:
        Transformed scalar field.
    """
    e = np.exp(- 4 * u / width)

    return amplitude * np.minimum(e, 1)


def hard_binarization(u: np.ndarray, threshold: float) -> np.ndarray:
    """
    Binarizes the Signed Distance field/pattern based on a threshold.
    Values below the threshold are 1 and values above are 0.
    
    Args:
        u: Signed Distance field or any scalar field.
        threshold: Binarization threshold.
    
    Returns:
        Binarized scalar field.
    """
    out = u <= threshold
    out = out.astype(float)
    return out


def linear_falloff(u: np.ndarray, amplitude: float | int, width: float | int) -> np.ndarray:
    """
    Applies a decreasing linear function to the scalar (Signed Distance Function) field.
    
    Args:
        u: Signed Distance field or any scalar field.
        amplitude: Maximum value of the transformed scalar field.
        width: Range at which the value of the transformed scalar field drops to zero.
    
    Returns:
        Transformed scalar field.
    """
    out = 1 - u/width

    return np.clip(out, 0, 1)*amplitude


def relu(u: np.ndarray, width: float | int = 1) -> np.ndarray:
    """
    Applies the ReLU function to the scalar (Signed Distance Function) field.
    
    Args:
        u: Signed Distance field or any scalar field.
        width: Range at which the value of the transformed field reaches one.
    
    Returns:
        Transformed scalar field.
    """
    return np.maximum(u/width, 0)


def smooth_relu(u: np.ndarray, smooth_width: float | int,
                width: float | int = 1, threshold: int | float = 0.01) -> np.ndarray:
    """
    Applies the "squareplus" function to the scalar (Signed Distance Function) field.
    https://en.wikipedia.org/wiki/Rectifier_(neural_networks)
    
    Args:
        u: Signed Distance field or any scalar field.
        smooth_width: Distance from the origin at which the Smooth ReLU function
            is greater than ReLU for less than the value of the threshold parameter.
        width: Range at which the value of the transformed field reaches one.
        threshold: At smooth_width distance from the origin the value of the Smooth ReLU function is greater
            than ReLU for the value of the threshold parameter.
            at smooth_width distance from the origin.
    
    Returns:
        Transformed scalar field.
    """
    b = (smooth_width + threshold)*4*threshold
    v = u/width
    return (v + np.sqrt(v**2 + b))/2


def slowstart(u: np.ndarray,
              smooth_width: float | int,
              width: float | int = 1, threshold: int | float = 0.01,
              ground: bool = True) -> np.ndarray:
    """
    Applies the SlowStart function to the scalar (Signed Distance Function) field.
    
    Args:
        u: Signed Distance field or any scalar field.
        smooth_width: Distance from the origin at which the SlowStart function
            is greater than ReLU for less than the value of the threshold parameter.
        width: Range at which the value of the transformed field reaches one.
        threshold: At smooth_width distance from the origin the value of the SlowStart function is greater
            than ReLU for the value of the threshold parameter.
        ground: if True the value of the function is zero at zero.
    
    Returns:
        Transformed scalar field.
    """
    b = (2*smooth_width + threshold)*threshold
    return np.sqrt(np.maximum(u / width, 0)**2 + b/width) - np.sqrt(b/width)*ground


def gaussian_boundary(u: np.ndarray, amplitude: float | int, width: float | int) -> np.ndarray:
    """
    Applies the Gaussian to the scalar (Signed Distance Function) field.
    
    Args:
        u: Signed Distance field or any scalar field.
        amplitude: Maximum value of the transformed scalar field.
        width: Range at which the value of the transformed scalar field drops to almost zero.
    
    Returns:
        Transformed scalar field.
    """
    out = np.exp(-4*(u/width)**2)

    return amplitude*out


def gaussian_falloff(u: np.ndarray, amplitude: float | int, width: float | int) -> np.ndarray:
    """
    Applies the Gaussian to the positive values of the scalar (Signed Distance Function) field.
    
    Args:
        u: Signed Distance field or any scalar field.
        amplitude: Maximum value of the transformed scalar field (and points at which the scalar field was < 0).
        width: Range at which the value of the transformed scalar field drops to almost zero.
    
    Returns:
        Transformed scalar field.
    """
    u = np.maximum(u, 0)
    out = np.exp(-4*(u/width)**2)

    return amplitude*out


def conv_averaging(u: np.ndarray, kernel_size: int | tuple | list | np.ndarray, iterations: int) -> np.ndarray:
    """
    Averages the field using an averaging convolutional kernel of the specified size.
    
    Args:
        u: Signed Distance field or any scalar field.
        kernel_size: Size of the averaging kernel. Must be an integer or a tuple/array of the
            same dimension as the scaler field.
        iterations: Number of times the convolutional averaging is applied to the input scalar field.
    
    Returns:
        Transformed scalar field.
    """
    if iterations == 0:
        return u

    shape_u = u.shape

    if isinstance(kernel_size, int):
        kernel_size = (kernel_size,)*len(shape_u)
    if isinstance(kernel_size, np.ndarray):
        kernel_size = tuple(kernel_size)
    if isinstance(kernel_size, list):
        kernel_size = tuple(kernel_size)

    if not len(kernel_size) == len(shape_u):
        raise ValueError("Dimension of the kernel and the field must match!")

    if len(kernel_size)==2:
        norm_ = kernel_size[0]*kernel_size[1]
    if len(kernel_size) == 3:
        norm_ = kernel_size[0] * kernel_size[1] * kernel_size[2]

    filter_ = np.ones(kernel_size) / norm_

    new = convolve(u, filter_)
    for i in range(iterations):
        new = convolve(new, filter_)

    return new


def conv_edge_detection(u: np.ndarray) -> np.ndarray:
    """
    Edge detection based on a 3x3 convolutional kernel.
    
    Args:
        u: Signed Distance field or any scalar field.
    
    Returns:
        Transformed scalar field.
    """

    if len(u.shape) == 2:
        filter_ = np.asarray([[-1, -1, -1], [-1, 8, -1], [-1, -1, -1]])
    if len(u.shape) == 3:
        f = np.asarray([[-1, -1, -1], [-1, 8, -1], [-1, -1, -1]])
        filter_ = np.zeros((3,3,1))
        filter_[:,:,:] = np.asarray([f]).T

    new = convolve(u, filter_)

    return new


def custom_post_process(u: np.ndarray,
                        function: Callable[[np.ndarray, tuple], np.ndarray],
                        parameters: tuple) -> np.ndarray:
    """
    Applies a custom user-specified post-processing function to a scalar (Signed Distance Function) field.
    
    Args:
        u: Signed Distance field or any scalar field.
        function: A custom post-processing function which takes the SDF as a first argument
            and the parameters of the function as the next arguments.
        parameters: Parameters of the custom post-processing function.
    
    Returns:
        Modified Scalar Field.
    """

    return function(u, *parameters)




# Copyright (C) 2024 Peter Ropač
# This file is part of SPOMSO.
# SPOMSO is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
# SPOMSO is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.
# You should have received a copy of the GNU Lesser General Public License along with SPOMSO. If not, see <https://www.gnu.org/licenses/>.

import numpy as np


def resolution_conversion(resolution: int) -> int:
    """
    Converts the given resolution so that there are odd number of points along each axis.

    Args:
        resolution: Given resolution along an axis.

    Returns:
        converted Resolution along an axis.
    """
    return int(resolution if resolution % 2 == 1 else resolution + 1)


def generate_grid(size: int | float | tuple | list | np.ndarray,
                  resolution: int | tuple | list | np.ndarray) -> (np.ndarray, tuple):
    """
    Generates a grid of points based on the provided size and resolution, centered at zero.
    The dimensionality of the grid is determined from the number of elements in the size input parameter.

    Args:
        size: size of the grid along each dimension.
        resolution: number of grid points along each dimension.

    Returns:
        Point cloud of points with shape (D, N),
                where D - dimensionality, N - total number of points in the grid.,
        Converted resolution of the grid, containing the number of points along each axis.
    """

    resolution = np.asarray(resolution)
    if resolution.size == 1:
        co_res_0 = resolution_conversion(resolution)
        co_res_1 = co_res_0
        co_res_2 = co_res_0

    if resolution.size == 2:
        co_res_0 = resolution_conversion(resolution[0])
        co_res_1 = resolution_conversion(resolution[1])
        co_res_2 = co_res_0

    if resolution.size == 3:
        co_res_0 = resolution_conversion(resolution[0])
        co_res_1 = resolution_conversion(resolution[1])
        co_res_2 = resolution_conversion(resolution[2])

    size = np.asarray(size)
    if size.size == 1:
        x = np.linspace(-size[0] / 2,
                        size[0] / 2,
                        co_res_0)
        coor = np.zeros((3, co_res_0))
        coor[0] = x

    if size.size == 2:
        x = np.linspace(-size[0] / 2,
                        size[0] / 2,
                        co_res_0)

        y = np.linspace(-size[1] / 2,
                        size[1] / 2,
                        co_res_1)

        co = np.asarray(np.meshgrid(x, y, indexing="ij"))
        co = co.reshape(2, -1)
        coor = np.zeros((3, co.shape[1]))
        coor[:2] = co

    if size.size == 3:
        x = np.linspace(-size[0] / 2,
                        size[0] / 2,
                        co_res_0)

        y = np.linspace(-size[1] / 2,
                        size[1] / 2,
                        co_res_1)

        z = np.linspace(-size[2] / 2,
                        size[2] / 2,
                        co_res_2)

        co = np.asarray(np.meshgrid(x, y, z, indexing="ij"))
        coor = co.reshape(3, -1)

    return coor, (co_res_0, co_res_1, co_res_2)


def smarter_reshape(pattern: np.ndarray, resolution: tuple | list | np.ndarray) -> np.ndarray:
    """
    Converts the Signed Distance field point cloud into a grid.

    Args:
        pattern: Signed Distance field
        resolution: Resolution of the grid, determining the number of points along each axis.

    Returns:
        Signed distance field in a grid.
    """

    n_ele = pattern.shape[0]
    resolution = np.asarray(resolution)

    if resolution.size == 1:
        res = resolution_conversion(resolution)
        if n_ele//res == 1:
            return pattern
        elif n_ele//(res**2) == 1:
            return pattern.reshape(res, res)
        elif n_ele//(res**3) == 1:
            return pattern.reshape(res, res, res)
        else:
            raise ValueError(f"Cannot reshape the pattern with shape {pattern.shape}")

    if resolution.size == 2:
        res0 = resolution_conversion(resolution[0])
        res1 = resolution_conversion(resolution[1])

        div = n_ele//(res0*res1)
        if div == 1:
            return pattern.reshape(res0, res1)

        elif not div == 1:
            if not div%1 == 0:
                raise ValueError(f"Cannot reshape the pattern with shape {pattern.shape}")
            else:
                return pattern.reshape(res0, res1, int(div))

        else:
            raise ValueError(f"Cannot reshape the pattern with shape {pattern.shape}")

    if resolution.size == 3:
        res0 = resolution_conversion(resolution[0])
        res1 = resolution_conversion(resolution[1])
        res2 = resolution_conversion(resolution[2])

        div = n_ele // (res0 * res1 * res2)
        if div==1:
            return pattern.reshape(res0, res1, res2)
        else:
            raise ValueError(f"Cannot reshape the pattern with shape {pattern.shape}")


def vector_smarter_reshape(pattern: np.ndarray, resolution: tuple | list | np.ndarray) -> np.ndarray:
    """
    Converts a vector field point cloud into a grid.

    Args:
        pattern: Vector field
        resolution: Resolution of the grid, determining the number of points along each axis.

    Returns:
        Vector field on a grid of shape (3, resolution[0], resolution[1], resolution[2]).
    """

    x = smarter_reshape(pattern[0], resolution)
    y = smarter_reshape(pattern[1], resolution)
    z = smarter_reshape(pattern[2], resolution)

    return np.asarray([x, y, z])


def nd_vector_smarter_reshape(pattern: np.ndarray, resolution: tuple | list | np.ndarray) -> np.ndarray:
    """
    Converts an n-dimensional vector field point cloud into a grid.

    Args:
        pattern: Vector field
        resolution: Resolution of the grid, determining the number of points along each axis.

    Returns:
        Vector field on a grid of shape (ND, resolution[0], resolution[1], resolution[2]).
    """

    c = smarter_reshape(pattern[0], resolution)
    out = np.zeros((pattern.shape[0], *c.shape))
    out[0] = c
    for i in range(1, pattern.shape[0]):
        out[i] = smarter_reshape(pattern[i], resolution)

    return out


def binning(pattern: np.ndarray, bins: int, equal_width: bool = True) -> np.ndarray:
    """
    Discretize the Signed Distance field/pattern to based on the specified number of bins.

    Args:
        pattern: Signed Distance field or any field.
        bins: Number of bins - unique discrete values in the final pattern.
        equal_width: All the bins are of equal width. If False the first and the last bin have half the width.

    Returns:
        Modified field.
    """

    max_ = np.amax(pattern)
    min_ = np.amin(pattern)
    a = (max_ - min_)
    v = (pattern - min_)/a

    if equal_width:
        u = (v * bins).astype(int) / (bins - 1)
    else:
        u = np.round(v * (bins - 1), 0) / (bins - 1)

    out = u*a + min_
    return out












